unit Intersections;

interface

uses
  Windows, SysUtils, FlexBase, FlexPath, FlexControls, FlexUtils;

type
  TIterateSegmentsProc = procedure(const p1, p2: TPoint;
    index1, index2: integer) of object;

  TFlexIntersectionControlInfo = record
   Control: TFlexControl;
   Points: TPointArray;
   Types: TPointTypeArray;
   PathInfo: TPathInfo;
  end;

  TFlexIntersections = class
  private
   FPoints: TPointArray;
   FCount: integer;
   FInfoA: TFlexIntersectionControlInfo;
   FInfoB: TFlexIntersectionControlInfo;
   FSegmentAPoint1: TPoint;
   FSegmentAPoint2: TPoint;
   function GetPoint(Index: integer): TPoint;
  protected
   procedure CheckGrow;
   procedure CompleteInfo(var AInfo: TFlexIntersectionControlInfo);
   procedure RebuildControlInfo(AControl: TFlexControl;
     var AInfo: TFlexIntersectionControlInfo); virtual;
   procedure ProcessASegment(const p1, p2: TPoint; Index1, Index2: integer);
   procedure ProcessBSegment(const p1, p2: TPoint; Index1, Index2: integer);
   function  AddIntersectionPoint(const p: TPoint): integer;
  public
   procedure SetControlA(Control: TFlexControl);
   procedure SetControlB(Control: TFlexControl);
   function  RebuildIntersectionPoints: boolean;
   property  PointCount: integer read FCount;
   property  Points[Index: integer]: TPoint read GetPoint;
  end;

procedure IterateFlatSegments(const Points: TPointArray;
  const Types: TPointTypeArray; const Info: TPathInfo;
  Proc: TIterateSegmentsProc);

implementation

procedure IterateFlatSegments(const Points: TPointArray;
  const Types: TPointTypeArray; const Info: TPathInfo;
  Proc: TIterateSegmentsProc);
var
  i, Index, Index2, Last: integer;
begin
  for i:=0 to Length(Info.Figures)-1 do begin
    Index := Info.Figures[i].FirstNode;
    Last := Info.Figures[i].LastNode;
    while Index <= Last do begin
      Index2 := Index + 1;
      if Index = Last then 
        if Types[Index] = ptEndNodeClose then
          Index2 := Info.Figures[i].FirstNode
        else
          break;
      Proc(Points[Index], Points[Index2], Index, Index2);
      inc(Index);
    end;
  end;
end;

// TFlexIntersections /////////////////////////////////////////////////////////

procedure TFlexIntersections.CheckGrow;
begin
 if FCount < Length(FPoints) then exit;
 SetLength(FPoints, FCount + 16);
end;

function TFlexIntersections.GetPoint(Index: integer): TPoint;
begin
 if (Index < 0) or (Index >= FCount) then
  raise Exception.CreateFmt('Point index out of bounds (%d)', [Index]);
 Result := FPoints[Index];
end;

procedure TFlexIntersections.SetControlA(Control: TFlexControl);
begin
 RebuildControlInfo(Control, FInfoA);
end;

procedure TFlexIntersections.SetControlB(Control: TFlexControl);
begin
 RebuildControlInfo(Control, FInfoB);
end;

function TFlexIntersections.AddIntersectionPoint(const p: TPoint): integer;
begin
 CheckGrow;
 Result := FCount;
 FPoints[Result] := p;
 inc(FCount);
end;

procedure TFlexIntersections.RebuildControlInfo(AControl: TFlexControl;
  var AInfo: TFlexIntersectionControlInfo);
begin
 with AInfo do begin
  // Initially reset info
  Control := Nil;
  Points := Nil;
  Types := Nil;
  PathInfo.PointCount := 0;
  PathInfo.IsCurve := False;
  PathInfo.Figures := Nil;
  if AControl is TFlexCurve then begin
   Control := AControl;
   TFlexCurve(Control).GetPointsEx(Points, Types);
   CompleteInfo(AInfo);
  end;
 end;
end;

procedure TFlexIntersections.CompleteInfo(
  var AInfo: TFlexIntersectionControlInfo);
var
  Curvature: double;

 procedure OffsetPath(var Points: TPointArray; OfsX, OfsY: integer);
 var
  i: integer;
 begin
  for i:=0 to Length(Points)-1 do begin
   inc(Points[i].X, OfsX);
   inc(Points[i].Y, OfsY);
  end;
 end;

begin
 // Here we must have filled Control, Points and Types fields in AInfo record
 with AInfo do begin
  if not Assigned(Control) then exit;
  // Flatten path (replace all bezier segments to polylines
  // Calc curvature accordingly to current scale
  Curvature := (Control.Owner.Scale/100) * (1/PixelScaleFactor);
  FlattenPath(Points, Types, Curvature);
  // Convert coords to document coords
  with Control.DocRect do
    OffsetPath(Points, Left, Top);
  // Store path info
  GetPathInfo(Points, Types, PathInfo);
 end;
end;

function TFlexIntersections.RebuildIntersectionPoints: boolean;
begin
 FCount := 0;
 Result := Assigned(FInfoA.Control) and Assigned(FInfoB.Control);
 if not Result then exit;
 // Iterate all segments for ControlA
 IterateFlatSegments(FInfoA.Points, FInfoA.Types, FInfoA.PathInfo,
   ProcessASegment);
end;

procedure TFlexIntersections.ProcessASegment(const p1, p2: TPoint;
  Index1, Index2: integer);
begin
 // Store segment info
 FSegmentAPoint1 := p1;
 FSegmentAPoint2 := p2;
 // Iterate all segments for ControlB
 IterateFlatSegments(FInfoB.Points, FInfoB.Types, FInfoB.PathInfo,
   ProcessBSegment);
end;

procedure TFlexIntersections.ProcessBSegment(const p1, p2: TPoint;
  Index1, Index2: integer);
var
  c1, c2: double;
  ip: TPoint;
begin
 try
  // Check intersection
  c1 :=
    ( 1.0 * (FSegmentAPoint1.Y - p1.Y) * (p2.X - p1.X) -
      1.0 * (FSegmentAPoint1.X - p1.X) * (p2.Y - p1.Y) ) /
    ( 1.0 * (FSegmentAPoint2.X - FSegmentAPoint1.X) * (p2.Y - p1.Y) -
      1.0 * (FSegmentAPoint2.Y - FSegmentAPoint1.Y) * (p2.X - p1.X) );
  c2 :=
    ( 1.0 * (FSegmentAPoint1.Y - p1.Y) * (FSegmentAPoint2.X - FSegmentAPoint1.X) -
      1.0 * (FSegmentAPoint1.X - p1.X) * (FSegmentAPoint2.Y - FSegmentAPoint1.Y) ) /
    ( 1.0 * (FSegmentAPoint2.X - FSegmentAPoint1.X) * (p2.Y - p1.Y) -
      1.0 * (FSegmentAPoint2.Y - FSegmentAPoint1.Y) * (p2.X - p1.X) );
  if (c1 >= 0.0) and (c1 <= 1.0) and
     (c2 >= 0.0) and (c2 <= 1.0) then begin
    // We have intersection of this sigments
    // Calc intersection point
    ip.X := p1.x + Round( (p2.x - p1.x) * c2 );
    ip.Y := p1.y + Round( (p2.y - p1.y) * c2 );
    AddIntersectionPoint(ip);
  end;
 except
   // No intersection
   Exit;
 end;
end;

end.
