unit GraphControls;

interface

uses
  Windows, Graphics, Classes, SysUtils,
  FlexBase, FlexControls, FlexUtils, FlexProps;

const
  NodeDiameter = 20; // pixels

  // Node colors
  NodeNormal              = clSilver;
  NodeHighlighted         = clYellow;
  NodeDisabled            = clRed;
  NodeMarker              = clLime;

  NodeOutlineNormal       = clBlack;
  NodeOutlineHighlighted  = clOlive;
  NodeOutlineMarker       = clGreen;

  // Line colors
  LineNormal              = clGreen;
  LineDesigning           = clNavy;
  LineForDelete           = clRed;

  LineWidth               = 2; // pixels
  LineForDeleteWidth      = 3; // pixels

type
  TGraphNode = class(TFlexEllipse)
  private
   FHighlighted: boolean;
   FDisabled: boolean;
   FOnMove: TNotifyEvent;
   FIsMarker: boolean;
   procedure SetDisabled(const Value: boolean);
   procedure SetHighlighted(const Value: boolean);
   function  GetCenter: TPoint;
   procedure SetCenter(const Value: TPoint);
   procedure SetIsMarker(const Value: boolean);
  protected
   procedure ControlCreate; override;
   procedure ControlDestroy; override;
   procedure PropChanged(Sender: TObject; Prop: TCustomProp); override;
   procedure UpdateLineLinks;
  public
   procedure Restyle;
   property  Highlighted: boolean read FHighlighted write SetHighlighted;
   property  Disabled: boolean read FDisabled write SetDisabled;
   property  IsMarker: boolean read FIsMarker write SetIsMarker;
   property  Center: TPoint read GetCenter write SetCenter;
   property  OnMove: TNotifyEvent read FOnMove write FOnMove;
  end;

  TGraphLine = class(TFlexCurve)
  private
   FNodeLinksProp: TStrProp;
   FDesigning: boolean;
   FForDelete: boolean;
   FOnMove: TNotifyEvent;
   FNodeA: TGraphNode;
   FNodeB: TGraphNode;
   procedure SetDesigning(const Value: boolean);
   procedure SetNodeA(const Value: TGraphNode);
   procedure SetNodeB(const Value: TGraphNode);
   procedure SetForDelete(const Value: boolean);
  protected
   procedure ControlCreate; override;
   procedure ControlDestroy; override;
   procedure CreateProperties; override;
   procedure PropChanged(Sender: TObject; Prop: TCustomProp); override;
   procedure PropStored(Sender: TObject; Prop: TCustomProp;
     var IsStored: boolean; const PropName: string = ''); override;
   procedure UpdateLinks;
   procedure NodeLinksLoaded;
  public
   procedure Restyle;
   procedure UpdatePoints;
   property  Designing: boolean read FDesigning write SetDesigning;
   property  ForDelete: boolean read FForDelete write SetForDelete;
   property  NodeA: TGraphNode read FNodeA write SetNodeA;
   property  NodeB: TGraphNode read FNodeB write SetNodeB;
   property  NodeLinksProp: TStrProp read FNodeLinksProp;
   property  OnMove: TNotifyEvent read FOnMove write FOnMove;
  end;

var
  GraphNodes: TList;
  GraphLines: TList;

implementation

{$IFDEF CONDITIONALEXPRESSIONS}
  {$IF CompilerVersion >= 26.0} // XE5
    {$LEGACYIFEND ON}
  {$IFEND}
  {$IF CompilerVersion >= 20.0} // Delphi/C++Builder 12.0 (Studio 2009)
uses Types;
  {$IFEND}
{$ENDIF}

// TGraphNode /////////////////////////////////////////////////////////////////

procedure TGraphNode.ControlCreate;
begin
 inherited;
 // Node diameter
 WidthProp.Value := ScalePixels(NodeDiameter);
 HeightProp.Value := WidthProp.Value;
 // Sets width & height props read only
 WidthProp.Style := WidthProp.Style + [ psReadOnly ];
 HeightProp.Style := HeightProp.Style + [ psReadOnly ];
 // Set default style
 Restyle;
 // Add to GraphNodes list
 if not Assigned(GraphNodes) then GraphNodes := TList.Create;
 GraphNodes.Add(Self);
end;

procedure TGraphNode.ControlDestroy;
var i: integer;
begin
 // Remove all line links to self 
 if Assigned(GraphLines) then
  for i:=0 to GraphLines.Count-1 do with TGraphLine(GraphLines[i]) do begin
   if Self = NodeA then NodeA := Nil;
   if Self = NodeB then NodeB := Nil;
  end;
 // Remove from GraphNodes list
 if Assigned(GraphNodes) then GraphNodes.Remove(Self);
 inherited;
end;

procedure TGraphNode.SetDisabled(const Value: boolean);
begin
 if Value = FDisabled then exit;
 FDisabled := Value;
 Restyle;
end;

procedure TGraphNode.SetHighlighted(const Value: boolean);
begin
 if Value = FHighlighted then exit;
 FHighlighted := Value;
 Restyle;
end;

procedure TGraphNode.SetIsMarker(const Value: boolean);
begin
 if Value = FIsMarker then exit;
 FIsMarker := Value;
 Restyle;
end;

procedure TGraphNode.Restyle;
begin
 if FIsMarker then begin
  BrushProp.Color := NodeMarker;
  PenProp.Color := NodeOutlineMarker;
 end else
 if FHighlighted then begin
  BrushProp.Color := NodeHighlighted;
  PenProp.Color := NodeOutlineHighlighted;
 end else begin
  // Normal outline
  PenProp.Color := NodeOutlineNormal;
  if FDisabled
   then BrushProp.Color := NodeDisabled
   else BrushProp.Color := NodeNormal;
 end;
end;

procedure TGraphNode.UpdateLineLinks;
var i: integer;
begin
 if not Assigned(GraphLines) then exit;
 for i:=0 to GraphLines.Count-1 do with TGraphLine(GraphLines[i]) do
  if (Self = NodeA) or (Self = NodeB) then UpdateLinks;
end;

procedure TGraphNode.PropChanged(Sender: TObject; Prop: TCustomProp);
var i: integer;
begin
 inherited;
 if Prop = IdProp then begin
  // Update all line links
  UpdateLineLinks;
 end else
 if ((Prop = WidthProp) or (Prop = HeightProp) or
     (Prop = LeftProp) or (Prop = TopProp)) then begin
  // Update all linked lines
  if Assigned(GraphLines) then begin
   for i:=0 to GraphLines.Count-1 do with TGraphLine(GraphLines[i]) do
    if (NodeA = Self) or (NodeB = Self) then UpdatePoints;
  end;
  // Generate move event
  if Assigned(FOnMove) then FOnMove(Self);
 end;
end;

function TGraphNode.GetCenter: TPoint;
var Doc: TRect;
begin
 Doc := DocRect;
 Result.x := Doc.Left + RectWidth(Doc) div 2;
 Result.y := Doc.Top + RectHeight(Doc) div 2;
end;

procedure TGraphNode.SetCenter(const Value: TPoint);
var NewCenter: TPoint;
    Control: TFlexControl;
begin
 NewCenter := Value;
 // Convert doc to parent coords
 Control := Parent;
 while Assigned(Control) do begin
  dec(NewCenter.x, Control.Left);
  dec(NewCenter.y, Control.Top);
  Control := Control.Parent;
 end;
 // Set new position
 LeftProp.Value := NewCenter.x - WidthProp.Value div 2;
 TopProp.Value := NewCenter.y - HeightProp.Value div 2;
end;

// TGraphLine /////////////////////////////////////////////////////////////////

procedure TGraphLine.ControlCreate;
begin
 inherited;

 // Set default (normal) style
 Restyle;
 // Add to GraphLines list
 if not Assigned(GraphLines) then GraphLines := TList.Create;
 GraphLines.Add(Self);
end;

procedure TGraphLine.CreateProperties;
begin
 inherited;
 // Create text-style node links (for store in files)
 FNodeLinksProp := TStrProp.Create(Props, 'NodeLinks');
 // Set "read-only" and "reference" (for resolve after whole document loaded)
 // style
 FNodeLinksProp.Style := FNodeLinksProp.Style + [ psReadOnly, psRef ];
end;

procedure TGraphLine.ControlDestroy;
begin
 // Remove from GraphLines list
 if Assigned(GraphLines) then GraphLines.Remove(Self);
 inherited;
end;

procedure TGraphLine.SetDesigning(const Value: boolean);
begin
 if Value = FDesigning then exit;
 FDesigning := Value;
 Restyle;
end;

procedure TGraphLine.SetForDelete(const Value: boolean);
begin
 if Value = FForDelete then exit;
 FForDelete := Value;
 Restyle;
end;

procedure TGraphLine.Restyle;
begin
 // Set pen color
 if FDesigning then
  PenProp.Color := LineDesigning
 else
 if FForDelete
  then PenProp.Color := LineForDelete
  else PenProp.Color := LineNormal;
 // Set pen width
 if FForDelete
  then PenProp.Width := ScalePixels(LineForDeleteWidth)
  else PenProp.Width := ScalePixels(LineWidth);
end;

procedure TGraphLine.PropChanged(Sender: TObject; Prop: TCustomProp);
begin
 inherited;
 if (Prop = FNodeLinksProp) and Props.Resolving then
  // Set links from loaded property
  NodeLinksLoaded
 else
 if Assigned(FOnMove) and
   ((Prop = WidthProp) or (Prop = HeightProp) or
    (Prop = LeftProp) or (Prop = TopProp)) then FOnMove(Self);
end;

procedure TGraphLine.PropStored(Sender: TObject; Prop: TCustomProp;
  var IsStored: boolean; const PropName: string);
begin
 if Prop = FNodeLinksProp then
  IsStored := Assigned(FNodeA) or Assigned(FNodeB)
 else
  inherited;
end;

procedure TGraphLine.UpdateLinks;
var TextLink: string;
begin
 // Check resolving (values from loaded FNodeLinksProp)
 if Props.Resolving then exit;
 // Create text link
 if Assigned(FNodeA)
  then TextLink := IntToStr(FNodeA.IdProp.Value)
  else TextLink := '0';
 TextLink := TextLink + ' - ';
 if Assigned(FNodeB)
  then TextLink := IntToStr(FNodeB.IdProp.Value)
  else TextLink := '0';
 // Set FNodeLinksProp
 FNodeLinksProp.Style := FNodeLinksProp.Style + [ psReadOnly ];
 try
  FNodeLinksProp.Value := TextLink;
 finally
  FNodeLinksProp.Style := FNodeLinksProp.Style + [ psReadOnly ];
 end;
 UpdatePoints;
end;

procedure TGraphLine.UpdatePoints;
begin
 // Update line points
 BeginUpdate;
 try
  with DocRect do begin
   if Assigned(FNodeA) then with NodeA.Center do
    Points[0] := Point(x - Left, y - Top);
   if Assigned(FNodeB) then with NodeB.Center do
    Points[1] := Point(x - Left, y - Top);
  end;
 finally
  EndUpdate;
 end;
end;

procedure TGraphLine.NodeLinksLoaded;
var Index, Cur: integer;
    NodeId: cardinal;
    TextLink: string;
begin
 // Extract linked nodes id's
 TextLink := FNodeLinksProp.Value;
 Index := Pos('-', TextLink);
 // Extract NodeA
 Cur := Index - 1;
 while (Cur > 1) and (TextLink[Cur] = ' ') do dec(Cur);
 NodeId := StrToIntDef(copy(TextLink, 1, Cur), 0);
 NodeA := TGraphNode(Owner.FindControlByID(NodeId));
 // Extract NodeB
 Cur := Index + 1;
 while (Cur < Length(TextLink)) and (TextLink[Cur] = ' ') do inc(Cur);
 NodeId := StrToIntDef(copy(TextLink, Cur, Length(TextLink)), 0);
 NodeB := TGraphNode(Owner.FindControlByID(NodeId));
end;

procedure TGraphLine.SetNodeA(const Value: TGraphNode);
begin
 if Value = FNodeA then exit;
 FNodeA := Value;
 UpdateLinks;
end;

procedure TGraphLine.SetNodeB(const Value: TGraphNode);
begin
 if Value = FNodeB then exit;
 FNodeB := Value;
 UpdateLinks;
end;

initialization
  RegisterFlexControl(TGraphNode);
  RegisterFlexControl(TGraphLine);

finalization
  FreeAndNil(GraphNodes);
  FreeAndNil(GraphLines);

end.
