unit FlexSatellite;

interface

uses
  Windows, Graphics,
  FlexBase, FlexControls, FlexProps, FlexUtils, FlexLibs;

const
  DefaultLinkRadius = 100000;

type
  TFlexOrbit = class(TFlexEllipse)
  protected
   procedure ControlCreate; override;
  public
   procedure SaveToFiler(Filer: TFlexFiler; const Indent: string); override;
  end;

  TFlexSatellite = class(TFlexEllipse)
  private
   FLinkProp: TLinkProp;
   FLinkRadiusProp: TIntProp;
   FLinkRadiusFixedProp: TBoolProp;
   FLinkAngleProp: TIntProp;
   FOrbitLinkProp: TLinkProp;
   FShowOrbitProp: TBoolProp;
   FInUpdatePosition: boolean;
  protected
   procedure CreateProperties; override;
   procedure ControlCreate; override;
   procedure PropChanged(Sender: TObject; Prop: TCustomProp); override;
   procedure PropStored(Sender: TObject; Prop: TCustomProp;
     var IsStored: boolean; const PropName: string = ''); override;
   procedure DoNotify(Notify: TFlexNotify); override;
   procedure LinkedNotify(Sender: TObject; Source: TNotifyLink;
     const Info: TNotifyLinkInfo);
   procedure UpdatePosition;
   procedure NormalizePosition;
  public
   property LinkProp: TLinkProp read FLinkProp;
   property LinkRadiusProp: TIntProp read FLinkRadiusProp;
   property LinkRadiusFixedProp: TBoolProp read FLinkRadiusFixedProp;
   property LinkAngleProp: TIntProp read FLinkAngleProp;
   property OrbitLinkProp: TLinkProp read FOrbitLinkProp;
   property ShowOrbitProp: TBoolProp read FShowOrbitProp;
  end;

implementation

// TFlexOrbit /////////////////////////////////////////////////////////////////

procedure TFlexOrbit.ControlCreate;
begin
 inherited;
 PenProp.Style := psDash;
 PenProp.Color := clSilver;
end;

procedure TFlexOrbit.SaveToFiler(Filer: TFlexFiler; const Indent: string);
begin
 // This control is never saves in filers
 Filer.Saved := Filer.Saved + 1;
end;

// TFlexSatellite /////////////////////////////////////////////////////////////

procedure TFlexSatellite.CreateProperties;
begin
 inherited;
 FLinkProp := TLinkProp.Create(Props, 'Link');
 FLinkProp.OnLinkedNotify := LinkedNotify;
 FLinkRadiusProp := TIntProp.Create(Props, 'LinkRadius');
 FLinkRadiusProp.Style := FLinkRadiusProp.Style + [psScalable];
 FLinkRadiusFixedProp := TBoolProp.Create(Props, 'LinkRadiusFixed');
 FLinkAngleProp := TIntProp.Create(Props, 'LinkAngle');
 FOrbitLinkProp := TLinkProp.Create(Props, 'OrbitLink');
 FOrbitLinkProp.Style := [psDontStore, psNonVisual, psRef];
 FShowOrbitProp := TBoolProp.Create(Props, 'ShowOrbit');
 FShowOrbitProp.Style := FShowOrbitProp.Style + [psNonVisual];
end;

procedure TFlexSatellite.ControlCreate;
begin
 inherited;
 FLinkRadiusProp.Value := DefaultLinkRadius;
 FLinkRadiusFixedProp.Value := true;
end;

procedure TFlexSatellite.DoNotify(Notify: TFlexNotify);
begin
 case Notify of
  fnLoaded,
  fnRect:
    if not FInUpdatePosition then NormalizePosition;
  fnOrder:
    if Assigned(FOrbitLinkProp.LinkedControl) and
      (FOrbitLinkProp.LinkedControl.ParentScheme <> ParentScheme) then begin
     // Move orbit with control
     FOrbitLinkProp.LinkedControl.Owner := Owner;
     FOrbitLinkProp.LinkedControl.Parent := ParentScheme;
    end;
 end;
 inherited;
end;

procedure TFlexSatellite.LinkedNotify(Sender: TObject; Source: TNotifyLink;
  const Info: TNotifyLinkInfo);
begin
 case Info.Code of
  ncControlNotify:
     if Info.ControlNotify = fnRect then begin
      UpdatePosition;
     end;
 end;
end;

procedure TFlexSatellite.PropChanged(Sender: TObject; Prop: TCustomProp);
var Orbit: TFlexEllipse;
begin
 if (Prop = FLinkProp) or
    (Prop = FLinkRadiusProp) or (Prop = FLinkAngleProp) then begin
  if Assigned(FLinkProp.LinkedControl) then UpdatePosition;
 end else
 if Prop = FOrbitLinkProp then begin
  if not Assigned(FOrbitLinkProp.LinkedControl) and
     not (fsDestroying in State) then
   FShowOrbitProp.Value := false;
 end else
 if Prop = FShowOrbitProp then begin
  if FShowOrbitProp.Value then begin
   // Create and link orbit control
   if not Assigned(FOrbitLinkProp.LinkedControl) and
      not (ParentScheme is TFlexLibrary) then begin
    Orbit := TFlexOrbit.Create(Owner, ParentScheme, Layer);
    Orbit.ToBack;
    FOrbitLinkProp.LinkedControl := Orbit;
    UpdatePosition;
   end;
  end else
  if Assigned(FOrbitLinkProp.LinkedControl) then begin
   // Destroy orbit control
   FOrbitLinkProp.LinkedControl.Free;
  end;
 end;
 inherited;
 if (Prop = LayerProp) and Assigned(FOrbitLinkProp.LinkedControl) then
  FOrbitLinkProp.LinkedControl.Layer := Layer;
end;

procedure TFlexSatellite.PropStored(Sender: TObject; Prop: TCustomProp;
  var IsStored: boolean; const PropName: string);
begin
 if Prop = FLinkProp then
  IsStored := Assigned(FLinkProp.LinkedControl)
 else
 if Prop = FLinkRadiusProp then
  IsStored := FLinkRadiusProp.Value <> DefaultLinkRadius
 else
 if Prop = FLinkRadiusFixedProp then
  IsStored := not FLinkRadiusFixedProp.Value
 else
 if Prop = FLinkAngleProp then
  IsStored := FLinkAngleProp.Value <> 0
 else
 if Prop = FOrbitLinkProp then
  IsStored := false
 else
 if Prop = FShowOrbitProp then
  IsStored := FShowOrbitProp.Value
 else
  inherited;
end;

procedure TFlexSatellite.UpdatePosition;
var MasterPos: TPoint;
    SelfPos: TPoint;
    RadAngle: double;
    Control: TFlexControl;
    Orbit: TFlexOrbit;
begin
 if FInUpdatePosition then exit;
 FInUpdatePosition := true;
 try
  if Assigned(FLinkProp.LinkedControl) then begin
   // Get master anchor position (center)
   with FLinkProp.LinkedControl.DocRect do begin
    MasterPos.X := (Left + Right) div 2;
    MasterPos.Y := (Top + Bottom) div 2;
   end;
   // Calc self position
   RadAngle := FLinkAngleProp.Value * pi / 180;
   SelfPos.X := MasterPos.X + Round(FLinkRadiusProp.Value * cos(RadAngle))
     -WidthProp.Value div 2;
   SelfPos.Y := MasterPos.Y - Round(FLinkRadiusProp.Value * sin(RadAngle))
     -HeightProp.Value div 2;
   // Convert to parent coordinates (if satellite grouped)
   Control := Parent;
   while Assigned(Control) do begin
    dec(SelfPos.X, Control.Left);
    dec(SelfPos.Y, Control.Top);
    Control := Control.Parent;
   end;
   // Set self position
   BeginUpdate;
   try
    Left := SelfPos.X;
    Top := SelfPos.Y;
   finally
    EndUpdate;
   end;
  end;
  // Update orbit
  Orbit := TFlexOrbit(FOrbitLinkProp.LinkedControl);
  if Assigned(Orbit) then begin
   if Assigned(FLinkProp.LinkedControl) then begin
    SelfPos.X := MasterPos.X - FLinkRadiusProp.Value;
    SelfPos.Y := MasterPos.Y - FLinkRadiusProp.Value;
    // Convert to parent coordinates (if orbit grouped)
    Control := Orbit.Parent;
    while Assigned(Control) do begin
     dec(SelfPos.X, Control.Left);
     dec(SelfPos.Y, Control.Top);
     Control := Control.Parent;
    end;
    // Set orbit size and position
    with Orbit do begin
     BeginUpdate;
     try
      Left := SelfPos.X;
      Top := SelfPos.Y;
      Width := 2 * FLinkRadiusProp.Value;
      Height := 2 * FLinkRadiusProp.Value;
      VisibleProp.Value := true;
     finally
      EndUpdate;
     end;
    end;
   end else begin
    // Satellite not linked. Orbit is invisible
    Orbit.VisibleProp.Value := false;
   end;
  end;
 finally
  FInUpdatePosition := false;
 end;
end;

procedure TFlexSatellite.NormalizePosition;
var MasterPos: TPoint;
    SelfPos: TPoint;
    NewRadius: integer;
    NewAngle: integer;
    DistX, DistY: double;
begin
 if not Assigned(FLinkProp.LinkedControl) then exit;
 with DocRect do begin
  SelfPos.X := (Left + Right) div 2;
  SelfPos.Y := (Top + Bottom) div 2;
 end;
 with FLinkProp.LinkedControl.DocRect do begin
  MasterPos.X := (Left + Right) div 2;
  MasterPos.Y := (Top + Bottom) div 2;
 end;
 DistX := SelfPos.X - MasterPos.X;
 DistY := -(SelfPos.Y - MasterPos.Y);
 FInUpdatePosition := true;
 try
  if not FLinkRadiusFixedProp.Value then begin
   NewRadius := Round(Sqrt(DistX * DistX + DistY * DistY));
   FLinkRadiusProp.Value := NewRadius;
  end;
  if DistX = 0 then begin
   if DistY > 0
    then NewAngle := 270
    else NewAngle := 90;
  end else
   NewAngle := Round(arctan(DistY / DistX) * 180 / pi);
  if DistX <= 0 then NewAngle := 180 + NewAngle;
  FLinkAngleProp.Value := NewAngle;
 finally
  FInUpdatePosition := false;
 end;
 UpdatePosition;
end;

initialization
  RegisterFlexControl(TFlexSatellite);
  // Line below is not necessary since TFlexOrbit controls never save
  // and accordingly never loads.
  // RegisterFlexControl(TFlexOrbit);
end.
